
open Logcache

module NodeId =
  struct
    type t = int
    let compare = Pervasives.compare
  end
module View = Set.Make(NodeId)

module Make (Log : Logic.T) : T =
  struct
    include Log

    let entails f g = Common.prof "cGeneric.entails" (fun () -> entails f g)

    let string_f f = Syntax.stringizer (print f)

    type elt = {
	mutable id : int;  (* node ident *)
	mutable f : t;  (* formula describing the node, and description of local object if some *)
	mutable obj : int;  (* number of objects that has this node as a descriptor *)
	mutable extpart : Ext.t;  (* object idents of the extent *)
	mutable sup : elt LSet.t;  (* list of parent nodes *)
	mutable sub : elt LSet.t;  (* list of child nodes *)
	mutable subpat : elt LSet.t; (* list of child nodes that are not values *)
	mutable shown : bool; (* whether has been shown manually *)
      }

  (* the node representing common start point for traversing functions *)
    let start : elt = {
      id = 0;
      f = top ();
      obj = 0;
      extpart = Ext.empty;
      sup = LSet.empty();
      sub = LSet.empty();
      subpat = LSet.empty ();
      shown = false;
    }

    let from_sub subs = 
      let subs =
	List.fold_left
	  (fun res n ->
	    if n.id = 0
	    then LSet.union res n.sub
	    else LSet.add n res)
	  (LSet.empty ())
	  subs in
      {start with sub = subs}

    let from_sup sups =
      let sups =
	List.fold_left
	  (fun res n ->
	    if n.id = 0
	    then LSet.union res n.sup
	    else LSet.add n res)
	  (LSet.empty ())
	  sups in
      {start with sup = sups}

  (* type of marks on nodes *)
    type mark = Unknown | Good | Bad

  (* type of views *)
    type view_type = Ideal of elt | Filter of elt

    type data = int * (oidpart, elt LSet.t) Hashtbl.t * (Term.id, elt LSet.t) Hashtbl.t * elt

    module Hashtbl_f =
      Hashtbl.Make
	(struct
	  type t = Log.t
	  let equal = (=)
	  let hash f = Hashtbl.hash (string_f f)
	end)

    class c (kernel : kernel) l =
      object (self)
	inherit [t,elt,data] default kernel l

      (* data structures *)
      (* --------------- *)

	val mutable cpt : int = 2 (* counter for new node idents : the current value is the ident of the next node *)
	val tbl_f : elt Hashtbl_f.t = Hashtbl_f.create 1001 (* hashtable for locating already existing nodes from syntax of formulas *)
	val mutable tbl_obj : (oidpart, elt LSet.t) Hashtbl.t = Hashtbl.create 1000 (* hashtable for locating objects among nodes *)
	val mutable tbl_term : (Term.id, elt LSet.t) Hashtbl.t = Hashtbl.create 1000 (* hashtable for locating nodes whose formula contains some term *)
	val weak_views : (view_type * View.t ref) Weak.t ref = ref (Weak.create 10) (* weak array of views used in ideals and filters *)
	val mutable hidden : t list = []
	val mutable v_top : elt = {
	  id = 1;
	  f = top ();
	  obj = 0;
	  extpart = Ext.empty;
	  sup = LSet.empty();
	  sub = LSet.empty();
	  subpat = LSet.empty ();
	  shown = false;
	}

      (* load and save methods *)
      (* --------------------- *)

	method save = (cpt, tbl_obj, tbl_term, v_top)

	method load (c, o, term, t) =
	  cpt <- c;
	  tbl_obj <- o;
	  tbl_term <- term;
	  v_top <- t;
	  weak_views := Weak.create 10;
	  Hashtbl_f.clear tbl_f;
	  self # fold Down
	    (fun n () -> Hashtbl_f.add tbl_f n.f n)
	    [v_top]
	    ()

      (* private methods *)
      (* --------------- *)

      (* recursive traversing of the graph *)
	method private fold_gen : 'a . elt -> (elt -> 'a -> 'a * elt list) -> ?f_bound:(elt -> 'a -> 'a) -> (elt -> 'a -> 'a) -> 'a -> 'a =
	  fun start f_begin_next ?f_bound f_end init ->
	    let m = Hashtbl.create 13
	    in self # fold_gen2 m start f_begin_next ?f_bound f_end init
	method private fold_gen2 : 'a . (int,bool) Hashtbl.t -> elt -> (elt -> 'a -> 'a * elt list) -> ?f_bound:(elt -> 'a -> 'a) -> (elt -> 'a -> 'a) -> 'a -> 'a =
	  fun m node f_begin_next ?f_bound f_end init ->
	    let id = node.id in
	    if not (Hashtbl.mem m id)
	    then begin
	      Hashtbl.add m id true;
	      let res0, ns = f_begin_next node init in f_end node
		(List.fold_left
		   (fun res n -> self # fold_gen2 m n f_begin_next ?f_bound f_end res)
		   res0
		   ns
		) end
	    else 
	      match f_bound with
	      | None -> init
	      | Some f -> f node init

      (* topologic traversing of the graph *)
	method private fold_topo : 'a 'b . elt -> ?prev:(elt -> int) -> (elt -> 'a -> 'a * 'b) -> (elt -> 'b option -> elt list) -> (elt -> 'a -> 'a) -> 'a -> 'a =
	  fun start ?prev f_begin next f_end init ->
	    let nprev = Hashtbl.create 13 in
	    if prev = None then (* initialize the number of previous nodes *)
	      self # fold_gen
		start
		(fun n () -> Hashtbl.add nprev n.id 1, next n None)
		~f_bound:(fun n () -> Hashtbl.replace nprev n.id (Hashtbl.find nprev n.id + 1))
		(fun n () -> ())
		();
	    self # fold_topo2
	      nprev
	      start
	      (match prev with Some p -> p | None -> fun n -> assert false)
	      (fun n res -> let res', y = f_begin n res in res', next n (Some y))
	      f_end
	      init
	method private fold_topo2 : 'a . (int,int) Hashtbl.t -> elt -> (elt -> int) -> (elt -> 'a -> 'a * elt list) -> (elt -> 'a -> 'a) -> 'a -> 'a =
	  fun nprev node prev f_begin_next f_end init ->
	    let id = node.id in
	    let np =
	      if not (Hashtbl.mem nprev id)
	      then
		let x = max 0 (prev node - 1) in
		Hashtbl.add nprev id x;
		x
	      else
		let x = Hashtbl.find nprev id - 1 in
		Hashtbl.replace nprev id x;
		x in
	    if np <= 0 (* last visit of node *)
	    then begin
	      assert (np >= -1);
	      Hashtbl.replace nprev id (-2);
	      let res0, ns = f_begin_next node init in
	      f_end node
		(List.fold_left
		   (fun res n -> self # fold_topo2 nprev n prev f_begin_next f_end res)
		   res0
		   ns) end
	    else (* wait for all previous nodes visited *)
	      init

      (* topological search for minimal elements *)
	method private first_topo : 'a 'b . ?prev:(elt -> int) -> pred:(elt -> 'a option * 'b) -> next:(elt -> 'b option -> elt list) -> start:elt -> 'a list =
	  fun ?prev ~pred ~next ~start ->
	    self # fold_topo
	      start
	      ?prev
	      (fun node res ->
		let id = node.id in
		if id=0
		then res, next node None
		else
		  match pred node with
		  | None, y -> res, next node (Some y)  (* node is bad *)
		  | Some x, y -> x::res, [])   (* node is good, and hence first *)
	      (fun node -> function Some y -> y | None -> next node None)
	      (fun node res -> res)
	      []

      (* search for last elements satisfying the predicate *)
	method private last : 'a . pred_next:(elt -> 'a option * elt list) -> start:elt -> 'a list =
	  fun ~pred_next ~start ->
	    let m = Hashtbl.create 13 in
	    snd (self # new_last m pred_next start [])
	method private new_last : 'a . (int,bool) Hashtbl.t -> (elt -> 'a option * elt list) -> elt -> 'a list -> bool * 'a list =
	  fun m pred_next node res ->
	    let x, ns = pred_next node in
	    let last, res1 = List.fold_left
		(fun (last,res) n ->
		  try
		    if Hashtbl.find m n.id
		    then (false,res) (* node is not a last one *)
		    else (last,res) (* node may be a last one *)
		  with Not_found -> (* m.(n.id) = Unknown *)
		    let last',res' = self # new_last m pred_next n res in
		    last & last', res')
		(true,res)
		ns in
	    let id = node.id in
	    if not last
	    then begin Hashtbl.add m id true; (false,res1) end
	    else match x with
	    | None -> begin Hashtbl.add m id false; (true,res1) end
	    | Some v -> begin Hashtbl.add m id true; (false,v::res1) end

      (* get all greatest lower bounds of a list of nodes *)
	method private glbs : pred:(elt -> bool) -> elt list -> elt list =
	  fun ~pred -> function
	    | [] -> [v_top]
	    | [node] when pred node -> [node]
	    | nodes ->
		let m = Hashtbl.create 13 in
		let lnodes = List.length nodes in
		List.iter
		  (fun node ->
		    self # fold Down
		      (fun n res ->
			let v = try Hashtbl.find m n.id with Not_found -> 0 in
			Hashtbl.replace m n.id (v+1))
		      [node]
		      ())
		  nodes;
		self # first_topo
		  ~prev:(fun node ->
		    List.fold_left
		      (fun res n ->
			let v = try Hashtbl.find m n.id with _ -> 0 in
			if v >= 1 then res+1 else res)
		      0 node.sup)
		  ~pred:(fun node ->
		    let v = try Hashtbl.find m node.id with _ -> 0 in
		    (if v >= lnodes && pred node then Some node else None), ())
		  ~next:(fun node _ -> node.sub)
		  ~start:(from_sub nodes)

      (* get the lowest upper bounds satisfying pred of a list of nodes *)
        method private lubs : 'a . pred:(elt -> int -> 'a option) -> suppmin:int -> elt list -> 'a list =
	  fun ~pred ~suppmin -> function
	    | [] -> raise (Invalid_argument "lubs : the list is empty") 
	    | nodes ->
		let m = Array.make cpt 0 in
		let l = List.length nodes in
		let supp = min suppmin l in
		List.iter
		  (fun node ->
		    self # fold Up (fun n res -> m.(n.id) <- m.(n.id)+1) [node] ())
		  (if supp=l then List.tl nodes else nodes);
		let smin = if supp=l then supp-1 else supp in
		self # first_topo
		  ?prev:None
		  ~pred:(fun node ->
		    let s = m.(node.id) in
		    (if s >= smin then pred node s else None), ())
		  ~next:(fun node _ -> node.sup)
		  ~start:(if supp=l then List.hd nodes else from_sup nodes)

      (* connect a node given a formula *)
	method private connect_feat : bool -> elt -> t -> unit =
	  fun isval new_node lf ->
	    List.iter
	      (fun node ->
		node.sub <- LSet.add new_node (LSet.subtract node.sub new_node.sub);
		if not isval then node.subpat <- LSet.add new_node (LSet.subtract node.subpat new_node.subpat))
	      new_node.sup;
	    List.iter
	      (fun node -> node.sup <- LSet.add new_node (LSet.subtract node.sup new_node.sup))
	      new_node.sub;
	    Common.weak_iter
	      !weak_views
	      (fun (v,view_ref) ->
		match v with
		| Ideal node -> if entails new_node.f node.f then view_ref := View.add new_node.id !view_ref
		| Filter node -> if entails node.f new_node.f then view_ref := View.add new_node.id !view_ref);
	    let sup_terms =
	      List.fold_left
		(fun res n -> LSet.union res (terms n.f))
		(LSet.empty ())
		new_node.sup in
	    List.iter
	      (fun term ->
		try Hashtbl.replace tbl_term term (LSet.add new_node (LSet.subtract (Hashtbl.find tbl_term term) new_node.sub))
		with Not_found -> Hashtbl.add tbl_term term [new_node])
	      (LSet.subtract (terms lf) sup_terms);
	    Hashtbl_f.add tbl_f lf new_node

      (* find or create the node whose formula is equivalent to the given formula f *)
	method private alloc : t -> elt * bool =
	  fun lf -> Common.prof "cGeneric.alloc" (fun () ->
	    try
	      let n = Hashtbl_f.find tbl_f lf in
	      n, false
	    with Not_found ->
	      let sup = LSet.of_list (Common.prof "cGeneric.alloc.last" (fun () ->
		self # last
		  ~pred_next:(fun node ->
		    let f = node.f in
		    if entails lf f
		    then Some node, node.subpat
		    else None, [])
		  ~start:v_top)) in
	      match sup with
	      | [node] when entails node.f lf ->
		  node.f <- lf;  (* the old formula is replaced by the new equivalent one *)
		  node, false  (* the searched node already exists *)
	      | nodes ->  (* the searched node must be inserted *)
		  let isval = isvalue lf in
		  let sub, extpart =
		    if isval
		    then [], Ext.empty
		    else 
                      let sub = LSet.of_list (Common.prof "cGeneric.alloc.glbs" (fun () ->
			self # glbs (fun node -> entails node.f lf) nodes)) in
                      sub, Ext.union_r (List.map (fun n -> n.extpart) sub) in
		  let new_node = {
		    id = (cpt <- cpt+1; cpt-1);
		    f = lf;
		    obj = 0;
		    extpart = extpart;
		    sup = sup;
		    sub = sub;
	            subpat = List.filter (fun n -> not (isvalue n.f)) sub;
		    shown = false;
		  } in
		  Common.prof "cGeneric.alloc.connect_feat" (fun () -> self # connect_feat isval new_node lf);
		  new_node, true)

      (* disconnect a node *)
	method private disconnect_feat : elt -> unit =
	  fun node ->
            let lf = node.f in
(*	    let s = string_f lf in *)
            List.iter
	      (fun n ->
		n.sub <- LSet.remove node n.sub;
		n.subpat <- LSet.remove node n.subpat)
	      node.sup;
            List.iter (fun n -> n.sup <- LSet.remove node n.sup) node.sub;
	    let mup = Hashtbl.create 13 in
            self # fold Up
	      (fun n () -> Hashtbl.replace mup n.id ())
	      node.sub
	      ();
            let mdown = Hashtbl.create 13 in
            let tbl = Hashtbl.create (2 * LSet.cardinal node.sub) in
            List.iter (fun n -> Hashtbl.add tbl n.id node.sup) node.sub;
            List.iter
	      (fun n ->
		List.iter (fun n' -> Hashtbl.remove mdown n'.id) node.sub;
		self # fold_gen
		  n
		  (fun n' () -> (Hashtbl.replace mdown n'.id ()), if Hashtbl.mem mup n'.id then n'.sub else [])
		  (fun n' () -> ())
		  ();
		let sub_old, sub_new = List.partition (fun n' -> Hashtbl.mem mdown n'.id) node.sub in
		n.sub <- LSet.union n.sub sub_new;
		n.subpat <- LSet.union n.subpat (List.filter (fun n' -> not (isvalue n'.f)) sub_new);
		List.iter (fun n' -> Hashtbl.replace tbl n'.id (LSet.remove n (Hashtbl.find tbl n'.id))) sub_old)
	      node.sup;
            List.iter
	      (fun n ->
		n.sup <- LSet.union n.sup (Hashtbl.find tbl n.id))
	      node.sub;
	    Common.weak_iter
	      !weak_views
	      (fun (v,view_ref) -> view_ref := View.remove node.id !view_ref);
	    List.iter
	      (fun term ->
		let l = Hashtbl.find tbl_term term in
		if LSet.mem node l then
		  let sub_elts =
		    List.filter
		      (fun n -> LSet.mem term (terms n.f))
		      node.sub in
		  Hashtbl.replace tbl_term term (LSet.union sub_elts (LSet.remove node l)))
	      (terms lf);
	    Hashtbl_f.remove tbl_f lf

      (* disconnect the node, except if it is an entry point, or a node whose obj is (Some _), or top *)
	method private free : elt -> unit =
	  fun node -> Common.prof "cGeneric.free" (fun () ->
	    if node.id > v_top.id && node.obj = 0
	    then self # disconnect_feat node)

	method private temp_alloc : 'a . t -> (elt -> 'a) -> 'a =
	  fun f proc ->
	    let node, created = self # alloc f in
	    let res = proc node in
	    if created then self # free node;
	    res

      (* update a node after an axiom has been added *)
	method private update : elt -> unit =
	  fun node ->
	    let f0 = node.f in
	    let isval = isvalue f0 in
	    let sup' = LSet.of_list (
	      self # last  (* nodes to be added as parents of node *)
		~pred_next:(fun n ->
		  if n.id = node.id then None, []
		  else if entails f0 n.f then Some n, n.subpat
		  else None, [])
		~start:v_top) in
	    let old_sup = LSet.subtract node.sup sup' in
	    let new_sup = LSet.subtract sup' node.sup in
	    List.iter
	      (fun n ->
		n.sub <- LSet.remove node n.sub;
		n.subpat <- LSet.remove node n.subpat)
	      old_sup;
	    List.iter
	      (fun n ->
		n.sub <- LSet.add node (LSet.subtract n.sub node.sub);
		if not isval then n.subpat <- LSet.add node (LSet.subtract n.subpat node.subpat))
	      new_sup;
	    List.iter
	      (fun n -> n.sup <- LSet.add node (LSet.subtract n.sup sup'))
	      node.sub;
	    node.sup <- sup';
	    Common.weak_iter
	      !weak_views
	      (fun (v,view_ref) ->
		match v with
		| Ideal n ->
		    if entails node.f n.f
		    then view_ref := View.add node.id !view_ref
		    else view_ref := View.remove node.id !view_ref
		| Filter n -> ());
	    self # fold Up
	      (fun n () -> n.extpart <- Ext.union_r (Ext.diff n.extpart node.extpart :: List.map (fun x -> x.extpart) n.sub))
	      old_sup
	      ();
	    self # fold Up
	      (fun n () -> n.extpart <- Ext.union n.extpart node.extpart)
	      new_sup
	      ()


      (* public methods *)
      (* -------------- *)

	method formula e = e.f

	method mem f = Hashtbl_f.mem tbl_f f

	method find f = Hashtbl_f.find tbl_f f

	method show f =
	  let e, created = self # alloc f in
	  if created
	  then begin 
	    hidden <- LSet.remove e.f hidden;
	    e.shown <- true
	  end

	method hide e =
	  if e.obj = 0
	  then begin
	    hidden <- LSet.add e.f hidden;
	    self # free e
	  end
(*
	  if Ext.is_empty e.extpart
	  then self # free e     (* empty nodes are removed from the cache *)
	  else e.show <- ManHide (* others are hidden *)
*)

	method shown =
	  self # fold Down
	    (fun n res -> if n.shown then n::res else res)
	    [v_top]
	    []

	method hidden = hidden
(*
	  self # fold Down
	    (fun n res -> if n.show = ManHide then n.f :: res else res)
	    [v_top]
	    []
*)

	method top = v_top

	method max = [v_top]

	method children e = e.sub

	method parents e = e.sup

	method ideal e = (* BEWARE: may be very inefficient, better use entails directly *)
	  let view_ref =
	    ref (self # fold Down
		   (fun n res -> View.add n.id res)
		   [e]
		   View.empty) in
	  Common.weak_add weak_views (Ideal e, view_ref);
	  (fun e' -> View.mem e'.id !view_ref)

	method filter e =
	  let view_ref =
	    ref (self # fold Up
		   (fun n res -> View.add n.id res)
		   [e]
		   View.empty) in
	  Common.weak_add weak_views (Filter e, view_ref);
	  (fun e' -> View.mem e'.id !view_ref)

	method fold : 'a . fold_mode -> (elt -> 'a -> 'a) -> elt list -> 'a -> 'a =
	  fun mode f l res ->
	    let start = match mode with Down -> from_sub l | Up -> from_sup l in
	    let next = match mode with Down -> self # children | Up -> self # parents in
	    self # fold_gen
	      start
	      (fun node res -> f node res, next node)
	      (fun node res -> res)
	      res


	method links oid = try Hashtbl.find tbl_obj oid with Not_found -> []

	method link oid d = Common.prof "cGeneric.link" (fun () ->
	  let node, created = self # alloc d in
	  if created then hidden <- LSet.remove node.f hidden;
	  if not (Ext.mem oid node.extpart) then begin
	    List.iter
	      (fun (vis,x) ->
		let n, c = self # alloc x in
		if c && LSet.mem n.f hidden
		then self # free n)
(*
		  match n.sub with
		  | [n'] when Ext.cardinal n'.extpart = Ext.cardinal n.extpart ->
		      if not n.shown then self # free n (* n.show <- AutoHide *)
		  | _ ->
		      () (* if n.show = AutoHide then n.show <- AutoShow*) )
*)
	      (features d);
	    node.obj <- node.obj+1;
	    let rec fold_add_ext n =
	      if not (Ext.mem oid n.extpart) then begin
		n.extpart <- Ext.add oid n.extpart;
		List.iter fold_add_ext n.sup
	      end in
	    Common.prof "cGeneric.link.fold.Ext.add" (fun () -> fold_add_ext node);
	    Hashtbl.replace tbl_obj oid (LSet.add node (try Hashtbl.find tbl_obj oid with Not_found -> LSet.empty ()))
	  end;
	  node)

	method unlink oid f =
	  self # temp_alloc f
	    (fun node -> self # unlink_elt oid node)
	    
	method unlink_elt oid node =
	  let links = self # links oid in
	  let p =
	    let f_node = node.f in
	    (fun n -> entails n.f f_node) in
	  let out, keep = List.partition p links in
	  if out <> [] then begin
	    let keep = if keep = [] then [v_top] else keep in (* oid must not be removed *)
	    Hashtbl.replace tbl_obj oid keep;
	    List.iter (fun n -> n.obj <- n.obj-1) out;
	    let f = self # filter (from_sup keep) in
	    self # fold Up
	      (fun n () ->
		if f n (* n subsumes some node in keep *)
		then ()
		else n.extpart <- Ext.remove oid n.extpart)
	      out
	      ()
	  end

	method replace oid d d' =
	  self # temp_alloc d
	    (fun node ->
	      if Ext.mem oid node.extpart &&
		not (List.exists (fun n -> Ext.mem oid n.extpart) node.sub) (* oid has exactly the property d *)
	      then begin
		self # unlink_elt oid node;
		ignore (self # link oid d')
	      end)

	method remove oid =
	  let links = self # links oid in
	  Hashtbl.remove tbl_obj oid;
	  List.iter (fun n -> n.obj <- n.obj-1) links;
	  self # fold Up
	    (fun n () ->
	      n.extpart <- Ext.remove oid n.extpart)
	    links
	    ()

(*
	  let empty_nodes =
	    self # fold Up
	      (fun n res ->
		n.extpart <- Ext.remove oid n.extpart;
		if Ext.is_empty n.extpart then LSet.add n res else res)
	      out
	      (LSet.empty ()) in
	  let empty_nodes =
	    self # fold Up
	      (fun n res ->
		let b = Ext.is_empty n.extpart in
		n.extpart <- Ext.add oid n.extpart;
		if b then LSet.remove n res else res)
	      keep
	      empty_nodes in
	  ()
*)
(*	  List.iter (self # free) empty_nodes *)

	method axiom f g_opt =
	  let terms = axiom f g_opt in
	  let nodes =
	    List.fold_left
	      (fun res t -> LSet.union res (try Hashtbl.find tbl_term t with Not_found -> LSet.empty ()))
	      (LSet.empty ())
	      terms in
	  List.iter (self # update) nodes

	method extpart f = Common.prof "cGeneric.extent" (fun () ->
	  self # temp_alloc f
	    (fun node -> node.extpart))

	method extpart_elt node = node.extpart

	method int ext =
	  if Ext.is_empty ext
	  then raise Not_found
	  else
(* buggy for more than 1 oid
	    if false && l = 0
	    then 
	      self # lubs
		~pred:(fun n _ -> Some n)
		~suppmin:(Ext.cardinal ext)
		(Ext.map (fun oid -> from_sup (self # links oid)) ext)
	    else
*)
	      self # last
		~pred_next:(fun n ->
		  if Ext.subset ext (self # ext_elt n)
		  then (Some n, n.sub)
		  else (None, []))
		~start:v_top

	method increments ~suppmin ext start = Common.prof "cGeneric.increments" (fun () ->
	  let start = match start with None -> v_top | Some n -> n in
	  let inter_ext = match ext with None -> (fun ext_n -> ext_n) | Some ext -> (fun ext_n -> Ext.inter ext ext_n) in
(*
	  self # first_topo
	    ?prev:(if start.id = v_top.id then Some (fun n -> LSet.cardinal n.sup) else None)
	    ~pred:(fun n -> Common.prof "cGeneric.increments.intersections" (fun () ->
	      let inter = inter_ext (kernel # ext l n.extpart) in
	      let i = Ext.cardinal inter in
	      (if shown n && i >= suppmin
	      then Some (i,inter,n)
	      else None),
	      (Some i)))
	    ~next:(fun n -> function
	      | Some (Some i) -> if i >= suppmin then n.sub else []
	      | _ -> n.sub)
	    ~start:(from_sub start.sub)
*)
(* trying to avoid fold_topo *)
	  match start.sub with
	  | [n] when not n.shown && Ext.cardinal n.extpart = Ext.cardinal start.extpart ->
	      ( match self # increments ~suppmin ext (Some n) with
	      | [] ->
		  let inter = inter_ext (kernel # ext l n.extpart) in
		  let i = Ext.cardinal inter in
		  if i >= suppmin
		  then [(i,inter,n)]
		  else []
	      | l -> l)
	  | _ ->
	      List.fold_left
		(fun res n -> Common.prof "cGeneric.increments.intersections" (fun () ->
		  let inter = inter_ext (kernel # ext l n.extpart) in
		  let i = Ext.cardinal inter in
		  if i >= suppmin
		  then (i,inter,n)::res
		  else res))
		[]
		start.sub)

      end
  end
