
module type PARAM =
  sig
  end

module type T =
  sig
    module CRoot : Logcache.ROOT

    val cache : CRoot.c ref

    module Atom :
	sig
	  type t =
	    | Formula of CRoot.t
	    | Elt of CRoot.elt

	  val formula : t -> CRoot.t
	end

    module Role :
	sig
	  type t =
	    | Forward of Atom.t
	    | Backward of Atom.t
	end

    module Query :
	sig
	  type t

        (* features composing a query *)
	  val features : t -> t list * t list (* positive and negative features *)
        (* query top *)
	  val top : t
        (* get the extent of a query *)
	  val extent : t -> Logcache.Ext.t
        (* get the increments *)
	  val increment : t -> t -> int * Logcache.Ext.t * Logcache.Ext.t
	  val increments : suppmin:int -> t -> t -> Logcache.Ext.t * (int * Logcache.Ext.t * t * Logcache.Ext.t) list
	(* separates propositional parts, and conjunctive roles *)
	  type splitted = t * t list
	  val split : t -> splitted
	  val merge : splitted -> t
        (* gets another point of view on the query through a role *)
	  val mirror : splitted -> t -> splitted
        (* comparison *)
	  val compare : t -> t -> int
        (* parse a query *)
	  val parse : Syntax.t_stream -> t
        (* parse a query increment, i.e., a function building a query from a query *)
	  val parse_incr : Syntax.t_stream -> (t -> t)
        (* print a query *)
	  val print : t -> Syntax.t_list
	end

    module Rules :
	sig
	  type elt = {
	      prop : Query.t;
	      mutable propext : Logcache.Ext.t;
	      addup : Source.update option;
	      remup : Source.update option
	    }

	  val add : Query.t -> Source.update option -> Source.update option -> elt
	  val remove : elt -> unit
	  val fold : ('a -> elt -> 'a) -> 'a -> 'a
	  val get_updates : Logcache.oid -> (Logcache.oid -> unit) -> Source.update list
	end

      module Update :
	  sig
	    type t = Add of CRoot.t | Sub of CRoot.t | Replace of CRoot.t * CRoot.t | Seq of t * t | Nop

	    val of_query : Query.t -> t
	    val of_descr : CRoot.t list -> t
	    val to_descr : t -> CRoot.t list
	    val parse : Syntax.t_stream -> t
	    val print : t -> Syntax.t_list
	  end

    type data

    val init : unit -> unit
    val save : unit -> data
    val load : data -> unit
  end

module Make (Param : PARAM) (CFeat : Logcache.T) : T =
  struct
    module CRoot = Logcache.Root CFeat
    module Ext = Logcache.Ext

    let cache = ref (new CRoot.c)

    module Atom =
      struct
	type t =
	  | Formula of CRoot.t
	  | Elt of CRoot.elt

	let formula = function
	  | Formula f -> f
	  | Elt e -> !cache # formula e

	let feat a = CRoot.feat (formula a)
	let role a = CRoot.role (formula a)

	let elt = function
	  | Formula f -> !cache # find f
	  | Elt e -> e

	let ext = function
	  | Formula f -> !cache # ext f
	  | Elt e -> !cache # ext_elt e

	let increments ~suppmin ext_opt atom_opt =
	  try
	    let elt_opt = Common.map_option elt atom_opt in
	    !cache # increments ~suppmin ext_opt elt_opt
	  with Not_found -> []

	let forward x ext =
	  match x with
	  | Formula f -> !cache # forward f ext
	  | Elt e -> !cache # forward_elt e ext

	let forward_increments ~suppmin ext_opt atom_opt =
	  try
	    let elt_opt = Common.map_option elt atom_opt in
	    !cache # forward_increments ~suppmin ext_opt elt_opt
	  with Not_found -> []

	let backward x ext =
	  match x with
	  | Formula f -> !cache # backward f ext
	  | Elt e -> !cache # backward_elt e ext

	let backward_increments ~suppmin ext_opt atom_opt =
	  try
	    let elt_opt = Common.map_option elt atom_opt in
	    !cache # backward_increments ~suppmin ext_opt elt_opt
	  with Not_found -> []

	let compare a1 a2 =
	  CRoot.compare (formula a1) (formula a2)

	let parse = parser
	  | [<d = CRoot.parse>] ->
	      try
(*		!cache # show d; *) (* commented because produce too spurious features *)
		let e =  !cache # find d in (* may raise Not_found *)
		Elt e
	      with Not_found ->
		Formula d

	let print a = CRoot.print (formula a)

      end

    module Role =
      struct
	type t =
	  | Forward of Atom.t
	  | Backward of Atom.t

	let reverse = function
	  | Forward a -> Backward a
	  | Backward a -> Forward a

	let through r ext =
	  match r with
	  | Forward a -> Atom.forward a ext
	  | Backward a -> Atom.backward a ext

	let compare r1 r2 =
	  match r1, r2 with
	  | Forward a1, Forward a2 -> Atom.compare a1 a2
	  | Backward a1, Backward a2 -> Atom.compare a1 a2
	  | _ -> Pervasives.compare r1 r2
      end

    module Query =
      struct
(*
	module MyOption =
	  Option.Make
	    (struct
	      let toks_some = []
	      let toks_none = []
	      let toks_bot = [Token.Sharp]
	    end)
	module MyProd =
	  Prod.Make
	    (struct
	      let toks_pre = []
	      let toks_sep = []
	      let toks_suf = []
	      let toks_bot = [Sharp]
	    end)
	module MyProp =
	  Prop.Make
	    (struct
	      let toks_all = [Token.Ident "all"]
	      let toks_none = [Token.Ident "none"]
	      let toks_and = [Token.Ident "and"]
	      let toks_or = [Token.Ident "or"]
	      let toks_not = [Token.Ident "not"]
	      let toks_except = [Token.Ident "except"]
	      let toks_implies = [Token.Ident "implies"]
	      let desc_atom = true
	    end)
	module MyIset = Iset.Make
	module rec P = MyProd CRoot (MyOption Log)
	and Log = MyProp (MyIset P)
*)

	type t =
	  | Atom of Atom.t
	  | Role of Role.t * t
	  | All
	  | Non
	  | Not of t
	  | And of t * t
	  | Or of t * t
	  | If of t * t (* conclusion IF premise *)
		(* - cannot be used inside other constructors
		   - equivalent to AND for extent, but different w.r.t. increments *)

      (* returns the list of features (positive and negative) composing a query *)
	let rec features q =
	  match q with
	  | Atom _ -> [q], []
	  | Role (_,_) -> [q], []
	  | Not (Atom _ as q1) -> [], [q1]
	  | Not (Role (_,_) as q1) -> [], [q1]
	  | All -> [], []
	  | Non -> [], []
	  | Not q1 ->
	      let lp, ln = features q1 in
	      ln, lp
	  | And (q1,q2) ->
	      let lp1, ln1 = features q1 in
	      let lp2, ln2 = features q2 in
	      lp1 @ lp2, ln1 @ ln2
	  | Or (q1,q2) ->
	      let lp1, ln1 = features q1 in
	      let lp2, ln2 = features q2 in
	      lp1 @ lp2, ln1 @ ln2
	  | If (q1,q2) ->
	      let lp1, ln1 = features q1 in
	      let lp2, ln2 = features q2 in
	      lp1 @ lp2, ln1 @ ln2

      (* simplification of a formula : ex., And (All, X) -> X*)
	let simpl_and : t -> t -> t =
	  fun f1 f2 ->
	    match f1, f2 with
	    | All, _ -> f2
	    | _, All -> f1
	    | _, _ -> And (f1,f2)

	let simpl_or : t -> t -> t =
	  fun f1 f2 ->
	    match f1, f2 with
	    | All, _
	    | _, All -> All
	    | _, _ -> Or (f1,f2)

	let rec simpl : t -> t =
	  function
	    | Atom a -> Atom a
	    | Role (r,f1) -> Role (r, simpl f1)
	    | All -> All
	    | Non -> Non
	    | Not f1 -> (match simpl f1 with
	      | All -> Non
	      | Non -> All
	      | f1' -> Not f1')
	    | And (f1,f2) -> simpl_and (simpl f1) (simpl f2)
	    | Or (f1,f2) -> simpl_or (simpl f1) (simpl f2)
	    | If (f1,f2) -> If (simpl f1, simpl f2)

(*
      (* fold on propositions of nodes *)
	let rec fold ~all ~none ~n ~a ~o ~atom ~forward ~backward = function
	  | Atom x -> atom x
	  | Forward (x,q) -> forward x (fold ~all ~none ~n ~a ~o ~atom ~forward ~backward q)
	  | Backward (x,q) -> backward x (fold ~all ~none ~n ~a ~o ~atom ~forward ~backward q)
	  | All -> all
	  | Non -> none
	  | Not q -> n (fold ~all ~none ~n ~a ~o ~atom ~forward ~backward q)
	  | And (q,q') -> a (fold ~all ~none ~n ~a ~o ~atom ~forward ~backward q) (fold ~all ~none ~n ~a ~o ~atom ~forward ~backward q')
	  | Or (q,q') -> o (fold ~all ~none ~n ~a ~o ~atom ~forward ~backward q) (fold ~all ~none ~n ~a ~o ~atom ~forward ~backward q')

	let to_formula =
	  fold
	    ~all:Log.All
	    ~none:Log.Non
	    ~n:(fun f -> Log.Not f)
	    ~a:(fun f1 f2 -> Log.And (f1,f2))
	    ~o:(fun f1 f2 -> Log.Or (f1,f2))
	    ~atom:(fun a f1_opt ->
	      let fa = formula_of_atom a in
	      Log.Atom (P.Other (fa, f1_opt))) (* TO BE DONE *)
*)

      (* query top *)
	let top = All

      (* Get the extent of a query *)
	let rec extent f =
	  match f with
	  | Atom a -> Atom.ext a
	  | Role (Role.Forward a, f1) -> Atom.backward a (extent f1)
	  | Role (Role.Backward a, f1) -> Atom.forward a (extent f1)
	  | All -> extent_top ()
	  | Non -> Ext.empty
	  | Not f1 -> Ext.diff (extent_top ()) (extent f1)
	  | And (f1,f2) -> Ext.inter (extent f1) (extent f2)
	  | Or (f1,f2) -> Ext.union (extent f1) (extent f2)
	  | If (f1,f2) -> Ext.inter (extent f1) (extent f2)
	and extent_top () =
	  !cache # ext_elt (!cache # top)

      (* increments *)

	let rec increments0 ~suppmin ext_opt =
	  let inter_ext = match ext_opt with None -> (fun e -> e) | Some ext -> (fun e -> Ext.inter ext e) in
	  function
	  | All ->
	      List.map
		(fun (i,inter,x) -> (i,inter,Atom (Atom.Elt x)))
		(Atom.increments ~suppmin ext_opt None) @
	      List.map
		(fun (i,inter,x) -> (i,inter,Role (Role.Forward (Atom.Elt x), top)))
		(Atom.forward_increments ~suppmin ext_opt None) @
	      List.map
		(fun (i,inter,x) -> (i,inter,Role (Role.Backward (Atom.Elt x), top)))
		(Atom.backward_increments ~suppmin ext_opt None)
	  | Atom a ->
	      List.map
		(fun (i,inter,x) -> (i,inter,Atom (Atom.Elt x)))
		(Atom.increments ~suppmin ext_opt (Some a))
	  | Role (Role.Forward a, f1) ->
	      List.map
		(fun (i,inter,x) -> (i,inter,Role (Role.Forward (Atom.Elt x), top)))
		(Atom.forward_increments ~suppmin ext_opt (Some a)) @
	      let ext1 = Ext.inter (Atom.forward a (match ext_opt with None -> extent_top () | Some ext -> ext)) (extent f1) in
	      let l1 = increments0 ~suppmin (Some ext1) f1 in
	      Common.mapfilter
		(fun (i1,inter1,x1) ->
		  let inter = inter_ext (Atom.backward a inter1) in
		  let i = Ext.cardinal inter in
		  if i >= suppmin
		  then Some (i,inter,Role (Role.Forward a, x1))
		  else None)
		l1
	  | Role (Role.Backward a, f1) ->
	      List.map
		(fun (i,inter,x) -> (i,inter,Role (Role.Backward (Atom.Elt x), top)))
		(Atom.backward_increments ~suppmin ext_opt (Some a)) @
	      let ext1 = Ext.inter (Atom.backward a (match ext_opt with None -> extent_top () | Some ext -> ext)) (extent f1) in
	      let l1 = increments0 ~suppmin (Some ext1) f1 in
	      Common.mapfilter
		(fun (i1,inter1,x1) ->
		  let inter = inter_ext (Atom.forward a inter1) in
		  let i = Ext.cardinal inter in
		  if i >= suppmin
		  then Some (i,inter,Role (Role.Backward a, x1))
		  else None)
		l1
	  | _ -> []

	let increments ~suppmin q v =
	  match q with
	  | If (q1,q2) ->
	      let ext_c = extent q1 in
	      let ext_h_opt = match q2 with All -> None | _ -> Some (extent q2) in
	      let l = increments0 ~suppmin ext_h_opt v in
	      ext_c,
	      List.map
		(fun (i,inter,x) ->
		  let inter_c = Ext.inter inter ext_c in
		  let i_c = Ext.cardinal inter_c in
		  (i_c, inter_c, x, inter))
		l
	  | q ->
	      let ext_q = extent q in
	      let ext_q_opt = match q with All -> None | _ -> Some ext_q in
	      let l = increments0 ~suppmin ext_q_opt v in
	      ext_q,
	      List.map
		(fun (i,inter,x) -> (i, inter, x, extent x))
		l

	let increment q x =
	  match q with
	  | If (q1,q2) ->
	      let ext_c = extent q1 in
	      let ext_hx = Ext.inter (extent q2) (extent x) in
	      let inter = Ext.inter ext_c ext_hx in
	      let i = Ext.cardinal inter in
	      (i,inter,ext_hx)
	  | q ->
	      let ext_q = extent q in
	      let ext_x = extent x in
	      let inter = Ext.inter ext_q ext_x in
	      let i = Ext.cardinal inter in
	      (i,inter,ext_x)

      (* entailment test between 2 queries *)
(*	let entails q1 q2 = Log.entails (to_formula q1) (to_formula q2) *)

	let rec entails f1 f2 =
	  match f1, f2 with
	  | Atom a1, Atom a2 -> CRoot.entails (Atom.formula a1) (Atom.formula a2)
	  | Atom a1, Or (f2a,f2b) -> entails f1 f2a || entails f1 f2b
	  | Or (f1a,f1b), _ -> entails f1a f2 && entails f1b f2
	  | Not f1a, Not f2a -> entails f2a f1a
	  | _, _ -> false

	let contradicts f1 f2 =
	  match f1, f2 with
	  | Not f1a, f2a -> entails f2a f1a
	  | f1a, Not f2a -> entails f1a f2a
	  | _, _ -> false

      (* intelligent conjunctions, removing redundant, trivial and contradictory elements from the query. *)
	let rec conj_add_sub_aux f2 = function
	  | And (f1a,f1b) ->
	      let add1a, f1a' = conj_add_sub_aux f2 f1a in
	      let add1b, f1b' = conj_add_sub_aux f2 f1b in
	      add1a && add1b, simpl_and f1a' f1b'
	  | f1 ->
	      let b1 = entails f1 f2 in
	      let b2 = entails f2 f1 in
	      not (b1 && b2),
	      if b1 || b2 || contradicts f1 f2 then All else f1

	let conj_add_sub f1 f2 =
	  match f1 with
	  | If (f1a, f1b) ->
	      let add, f1b' = conj_add_sub_aux f2 f1b in
	      if add
	      then If (f1a, simpl_and f1b' f2)
	      else If (f1a, f1b')
	  | _ ->
	      let add, f1' = conj_add_sub_aux f2 f1 in
	      if add
	      then simpl_and f1' f2
	      else f1'

      (* splitting queries in proposition and roles *)
	type splitted = t * t list

	let rec split : t -> splitted =
	  fun f ->
	    match f with
	    | Atom _ -> f, []
	    | Role (_, _) -> All, [f]
	    | All -> All, []
	    | Non -> Non, []
	    | And (f1, f2) ->
		let p1, lr1 = split f1 in
		let p2, lr2 = split f2 in
		simpl_and p1 p2, lr1 @ lr2
	    | Or (_, _) -> f, []
	    | Not _ -> f, []
	    | If (f1, f2) ->
		let p2, lr2 = split f2 in
		And (f1,p2), lr2

	let merge : splitted -> t =
	  fun (p, lr) ->
	    List.fold_left
	      (fun f r -> And (f, r))
	      p lr

      (* traversing a role as a mirror to get another point of view *)
	let mirror : splitted -> t -> splitted =
	  fun (p, lr) -> function
	    | Role (r, f1) ->
		let r1 = Role (Role.reverse r, merge (p, lr)) in
		let p1, lr1 = split f1 in
		(p1, r1::lr1)
	    | _ -> (p, lr)


	let rec compare f1 f2 = (* partial definition for navigation features *)
	  match f1, f2 with
	  | Atom a1, Atom a2 ->
	      Atom.compare a1 a2
	  | Role (r1,g1), Role (r2,g2) ->
	      Common.compare_pair (Role.compare,compare) (r1,g1) (r2,g2)
	  | _ -> Pervasives.compare f1 f2

	let rec parse0 () = Syntax.parse_prop {
	  Syntax.all = "all";
	  Syntax.none = "none";
	  Syntax.n = "not";
	  Syntax.a = "and"; 
	  Syntax.o = "or"; 
	  Syntax.an = "except";
	  Syntax.taut = All;
	  Syntax.cont = Non;
	  Syntax.neg = (fun f1 -> Not f1);
	  Syntax.conj = (fun f1 f2 -> And (f1,f2));
	  Syntax.disj = (fun f1 f2 -> Or (f1,f2));
	  Syntax.atom = parser
	    | [<a = Atom.parse; str>] ->
		if not (Atom.role a)
		then Atom a
		else match str with parser
		| [<'Token.Minus; 'Token.GT; f = parse0 ()>] -> Role (Role.Forward a, f)
		| [<'Token.LT; 'Token.Minus; f = parse0 ()>] -> Role (Role.Backward a, f)
		| [<>] -> if Atom.feat a then Atom a else raise Stream.Failure
	}

	let parse_prop = parse0 ()

	let parse = parser
	  | [<f1 = parse_prop; str>] ->
	      match str with parser
	      | [<'Token.Ident "if"; f2 = parse_prop>] -> If (f1,f2)
	      | [<>] -> f1

	let parse_incr = parser
	  | [<'Token.Ident "if">] -> (fun f' -> If (f',All))
	  | [<'Token.Ident "unif">] -> (function If (f1',f2') -> simpl_and f1' f2' | f' -> f')
(*
	  | [<'Token.Ident "conclusion">] -> (function If (f1',_) -> f1' | f' -> f')
	  | [<'Token.Ident "premise">] -> (function If (_,f2') -> f2' | f' -> f')
*)
	  | [<'Token.Ident "compl">] -> (fun f' -> Not f')
	  | [<'Token.Ident "not"; f = parse>] -> (fun f' -> conj_add_sub f' (Not f))
	  | [<f = parse>] -> (fun f' -> conj_add_sub f' f)


	type spec_prop = {all : string; none : string; n : string; a : string; an : string; o : string}

	let prec_or = 3
	let prec_and = 2
	let prec_not = 1

	let rec print0 spec prec = function
	  | Atom a ->
	      Atom.print a
	  | Role (Role.Forward a, f1) ->
	      let toks = print0 spec prec_not f1 in
	      Atom.print a @ [Token.PP_tilda; Token.Minus; Token.GT; Token.PP_tilda] @ toks
	  | Role (Role.Backward a, f1) ->
	      let toks = print0 spec prec_not f1 in
	      Atom.print a @ [Token.PP_tilda; Token.LT; Token.Minus; Token.PP_tilda] @ toks
	  | All ->
	      [Token.Ident spec.all]
	  | Non ->
	      [Token.Ident spec.none]
(*
	  | And (f1,Not f2) ->
	      let toks1 = print0 spec prec_and f1 in
	      let toks2 = print0 spec prec_not f2 in
	      Syntax.add_par prec prec_and
		(toks1 @ [Token.PP_space; Token.Ident spec.an; Token.PP_tilda] @ toks2)
*)
	  | And (f1,f2) ->
	      let toks1 = print0 spec prec_and f1 in
	      let toks2 = print0 spec prec_and f2 in
	      Syntax.add_par prec prec_and
		(toks1 @ [Token.PP_space; Token.Ident spec.a; Token.PP_tilda] @ toks2)
	  | Or (f1,f2) ->
	      let toks1 = print0 spec prec_or f1 in
	      let toks2 = print0 spec prec_or f2 in
	      Syntax.add_par prec prec_or
		(toks1 @ [Token.PP_space; Token.Ident spec.o; Token.PP_tilda] @ toks2)
	  | Not f1 ->
	      let toks1 = print0 spec prec_not f1 in
	      Token.Ident spec.n :: Token.PP_tilda :: toks1
	  | _ -> assert false
(*
	and print_partial : t -> Syntax.t_list * (int * t) list =
	  let f_ref cpt spec prec q = incr cpt; [Token.Dollar; Token.Nat !cpt], [(!cpt, q)] in
	  function q ->
	    let cpt = ref 0 in
	    print_f {all = "all"; none = "none"; n = "not"; a = "and"; an = "except"; o = "or"} prec_and (f_ref cpt) q.f
*)

	let print_prop = print0 {all = "all"; none = "none"; n = "not"; a = "and"; an = "except"; o = "or"}

	let print : t -> Syntax.t_list =
	  function
	    | If (f1,f2) -> print_prop prec_and f1 @ Token.PP_tilda :: Token.Ident "if" :: Token.PP_tilda :: print_prop prec_and f2
	    | f -> print_prop prec_and f

      end

    module Rules =
      struct
	type elt = {
	    prop : Query.t;
	    mutable propext : Ext.t;
	    addup : Source.update option;
	    remup : Source.update option
	  }

	let l : elt list ref = ref []

      (* add an add_update to a node *)
	let add q addup remup =
	  let rule = {prop=q; propext=Query.extent q; addup=addup; remup=remup} in
	  l := rule::!l;
	  rule

      (* remove an update rule *)
	let remove rule =
	  l := List.filter ((!=) rule) !l

      (* iteration on all rules *)
	let fold f init = List.fold_left f init !l

      (* update the extent of rules *)
	let update () =
	  List.iter
	    (fun rule -> rule.propext <- Query.extent rule.prop)
	    !l

      (* return updates to be performed after some operation linking and unlinking properties to some object *)
	let get_updates : Logcache.oid -> (Logcache.oid -> unit) -> Source.update list =
	  fun oid p ->
	    let rules_in = List.filter (fun rule -> Ext.mem oid rule.propext) !l in
	    let rules_out = List.filter (fun rule -> not (Ext.mem oid rule.propext)) !l in
	    p oid;
	    update ();
	    let rem_updates =
	      Common.mapfilter
		(fun rule ->
		  if not (Ext.mem oid rule.propext)
		  then Common.map_option (Source.oid_update oid) rule.remup
		  else None)
		rules_in in
	    let add_updates =
	      Common.mapfilter
		(fun rule ->
		  if Ext.mem oid rule.propext
		  then Common.map_option (Source.oid_update oid) rule.addup
		  else None)
		rules_out in
            rem_updates @ add_updates
      end

    module Update =
      struct
	type t = Add of CRoot.t | Sub of CRoot.t | Replace of CRoot.t * CRoot.t | Seq of t * t | Nop
	  (* Add's arguments must be descriptors *)

	let of_descr d =
	  List.fold_left (fun res x -> Seq (res, Add x)) Nop d

	let rec to_descr = function
	  | Add x -> [x]
	  | Sub x -> []
	  | Replace (x1,x2) -> []
	  | Seq (u1,u2) -> to_descr u1 @ to_descr u2
	  | Nop -> []

	let rec of_query f =
	  match f with
	  | Query.Atom a ->
	      let f = Atom.formula a in
	      if CRoot.desc f
	      then Add f
	      else Nop
	  | Query.Not (Query.Atom a) ->
	      let f = Atom.formula a in
	      Sub f
	  | Query.Not (Query.Not q1) ->
              of_query q1
	  | Query.Not (Query.Or (q1,q2)) ->
	      of_query (Query.And (Query.Not q1, Query.Not q2))
	  | Query.And (q1,q2) ->
	      let u1 = of_query q1 in
	      let u2 = of_query q2 in
	      Seq (u1,u2)
	  | _ ->
	      Nop

	open Token

	let rec parse = parser
	  | [<u = parse_atom; ff = parse_suite>] -> ff u
	and parse_suite = parser
	  | [<'Ident "and"; u2 = parse>] -> (fun u1 -> Seq (u1,u2))
	  | [<>] -> (fun u -> u)
	and parse_atom = parser
	  | [<'Ident "all">] -> Nop
	  | [<'Ident "not"; f = CRoot.parse>] -> Sub f
	  | [<'Ident "replace"; d = CRoot.parse; 'Ident "by"; d' = CRoot.parse>] ->
	      if CRoot.desc d && CRoot.desc d'
	      then Replace (d,d')
	      else raise (Stream.Error "Some property is not a valid object descriptor")
	  | [<d = CRoot.parse>] ->
	      if CRoot.desc d
	      then Add d
	      else raise (Stream.Error "Some property is not a valid object descriptor")

	let rec print = function
	  | Add x -> CRoot.print x
	  | Sub x -> Ident "not" :: PP_tilda :: CRoot.print x
	  | Replace (x,x') -> Ident "replace" :: CRoot.print x @ Ident "by" :: CRoot.print x'
	  | Seq (u1,u2) ->
	      ( match print u1, print u2 with
	      | [Ident "all"], toks2 -> toks2
	      | toks1, [Ident "all"] -> toks1
	      | toks1, toks2 -> toks1 @ PP_tilda :: Ident "and" :: PP_tilda :: toks2)
	  | Nop -> [Ident "all"]
      end

    let init () =
      Term.init ();
      cache := new CRoot.c;
      Rules.l := []

    type data = Term.data * CRoot.data * Rules.elt list

    let save () = (Term.save (), !cache # save, !Rules.l)

    let load (t, d, r) =
      Term.load t;
      !cache # load d;
      Rules.l := r

  end
