(** Type and signature for sources. *)

open Token

type url = string

let rel dirname url =
  if url = dirname
  then ""
  else if Str.string_match (Str.regexp (Str.quote dirname ^ "[/\\\\]\\(.+\\)$")) url 0
  then Str.matched_group 1 url
  else url

let abs dirname url =
  if url = ""
  then dirname
  else
    let abs_url = Filename.concat dirname url in
    if Filename.is_implicit url && Sys.file_exists abs_url (* not a URL *)
    then abs_url
    else url

(** The type of object previews. *)
type preview =
  | String of string
  | Picture of string * string  (** filename, comment *)

let print_preview dirname = function
  | String s -> [ Token.String s ]
  | Picture (file,s) -> [ Term "Picture"; LeftPar; Token.String (rel dirname file); Comma; Token.String s; RightPar ]

(** the type of action arguments. *)
type action_arg = {
    mime : string; (** A mime type used as a key: e.g., image/jpeg, url/text/html. *)
    role : string; (** The role of the argument w.r.t. the object: e.g., "", "PDF". *)
    location : string; (** The location (filename or url) of the file of type [mime], and having role [role]. *)
  }

let print_arg_option arg : Syntax.t_list =
  [ PP_tilda; LeftAcc;
    Token.String arg.mime; Comma; PP_tilda;
    Token.String arg.role; Comma; PP_tilda;
    Token.String arg.location;
    RightAcc ]

let print_arg_options args : Syntax.t_list = List.fold_right (fun arg res -> print_arg_option arg @ res) args []

(** Options transmitted to sources for controlling what is generated. *)
type src_option = [
  | `Parts      (** Should proper parts of files be represented by objects. *)
  | `Recursive  (** Should directories be explored recursively. *)
  | `Suffixes of string list (** Filter on file suffixes. *)
]

let print_src_options options : Syntax.t_list =
  if options = []
  then []
  else
    LeftBra ::
    List.fold_left
      (fun res -> function
	| `Parts -> Term "Parts" :: res
	| `Recursive -> Term "Recursive" :: res
	| `Suffixes l -> Term "Suffixes" :: Token.String (String.concat "." l) :: res)
      [RightBra]
      options


(** Type of background updates, as generated by sources and some functions in {!Context.T}. *)
type update =
(* BEWARE: the order of constructors below is used for defining priority between updates *)
  | Axiom of Syntax.t_list * Syntax.t_list * url
	(** [Axiom (f,g,src)] states a new axiom [f] => [g] from source [src] (when [src] <> [""]). *)
  | Action of string * string * Syntax.t_list
	(** [Action (name,cmd,query)] creates a new action. *)
  | MakeExtr of preview * action_arg list * Syntax.t_list
        (** [MakeExtr (pre,args,d)] creates a new object (from no source) with preview [pre], action arguments [args], and initial description [d]. *)
  | Make of url * string * Syntax.t_list
	(** [Make (src,ref,d)] creates a new object from source [src] with reference [ref], and initial description [d]. *)
  | Init of url * url * src_option list
	(** [Init (src,url,options)] initializes a source at [url] generated by source [src], according to [options]. *)
  | Move of url * url
	(** [Move (src,url)] relocates the source [src] at [url]. *)
  | Remove of Logcache.oid
	(** [Remove oid] removes the object [oid]. *)
  | RemoveSrc of url * string
	(** [RemoveSrc (src,ref)] removes the object [ref] in source [src]. *)
  | Final of url
	(** [Final (src)] removes the source [src]. *)
  | FinalWll of string
        (** [FinalWll (wll)] removes the well [wll]. *)
(*
  | FinalRuleIntr of Syntax.t_list * Syntax.t_list
        (** [FinalRuleIntr (p,c)] removes the intrinsic move rule with premise [p] and conclusion [c]. *)
  | FinalRuleExtr of Syntax.t_list * Syntax.t_list
        (** [FinalRuleExtr (p,c)] removes the extrinsic move rule with premise [p] and conclusion [c]. *)
  | FinalRuleRem of Syntax.t_list
        (** [FinalRuleRem (p)] removes the remove rule on [p]. *)
*)
  | UpdateIntrSrc of url * string * Syntax.t_list
	(** [UpdateIntrSrc (src,ref,d)] updates the intrinsic description of object [ref] in source [src] with the formula [d]. *)
  | MoveIntr of Logcache.oid * Syntax.t_list
	(** [MoveIntr (oid,e)] changes the source of [oid] so as to force update [e] on the intrinsic description. *)
  | MoveExtr of Logcache.oid * Syntax.t_list
	(** [MoveExtr (oid,e)] updates the extrinsic description of [oid] with the formula [e]. *)
  | MoveExtrSrc of url * string * Syntax.t_list
	(** [MoveExtrSrc (src,ref,e)] updates the extrinsic description of object [ref] in source [src] with the formula [e]. *)
  | InitRuleIntr of Syntax.t_list * Syntax.t_list
        (** [InitRuleIntr (p,c)] creates a intrinsic move rule with premise [p] and conclusion [c]. *)
  | InitRuleExtr of Syntax.t_list * Syntax.t_list
        (** [InitRuleExtr (p,c)] creates a extrinsic move rule with premise [p] and conclusion [c]. *)
  | InitRuleRem of Syntax.t_list
        (** [InitRuleRem (p)] creates a remove rule on [p]. *)
  | Show of Syntax.t_list
	(** [Show f] makes the feature [f] visible in navigation. *)
  | Hide of Syntax.t_list
	(** [Hide f] makes the feature [f] invisible in navigation. *)
  | InitWll of string * Syntax.t_list
        (** [InitWll (wll,q)] initializes a well [wll] on query [q]. *)
  | MoveWll of url * url
	(** [Move (wll,url)] relocates the well [wll] at [url]. *)
  | Update of url * src_option list
	(** [Update (src,options)] updates the source [src] according to [options]. *)
  | UpdateWll of string
	(** [UpdateWll (wll)] updates the well [wll]. *)

let print_update : update -> Syntax.t_list =
  function
    | Init (url_source,url,options) ->
	Term "Init" :: LeftPar :: Token.String url_source :: Comma :: Token.String url :: Comma :: print_src_options options @ [RightPar]
    | Update (url,options) ->
	Term "Update" :: LeftPar :: Token.String url :: Comma :: print_src_options options @ [RightPar]
    | Final (url) ->
	Term "Final" :: LeftPar :: Token.String url :: [RightPar]
    | Move (url,new_url) ->
	Term "Move" :: LeftPar :: Token.String url :: Comma :: Token.String new_url :: [RightPar]
    | InitWll (url,query) ->
	Term "InitWll" :: LeftPar :: Token.String url :: Comma :: query @ [RightPar]
    | UpdateWll (url) ->
	Term "UpdateWll" :: LeftPar :: Token.String url :: [RightPar]
    | FinalWll (url) ->
	Term "FinalWll" :: LeftPar :: Token.String url :: [RightPar]
    | MoveWll (url,new_url) ->
	Term "MoveWll" :: LeftPar :: Token.String url :: Comma :: Token.String new_url :: [RightPar]
    | Make (url,r,d) ->
	Term "Make" :: LeftPar :: Token.String url :: Comma :: Token.String r :: Comma :: d @ [RightPar]
    | UpdateIntrSrc (src,ref,d) ->
	Term "UpdateIntr" :: LeftPar :: Token.String src :: Comma :: Token.String ref :: Comma :: d @ [RightPar]
    | MakeExtr (pre,args,d) ->
	Term "MakeExtr" :: LeftPar :: print_preview "" pre @ Comma :: (print_arg_options args : Syntax.t_list) @ Comma :: d @ [RightPar]
    | MoveIntr (oid,u) ->
	Term "MoveIntr" :: LeftPar :: Nat oid :: Comma :: u @ [RightPar]
    | MoveExtr (oid,u) ->
	Term "MoveExtr" :: LeftPar :: Nat oid :: Comma :: u @ [RightPar]
    | MoveExtrSrc(url,ref,u) ->
	Term "MoveExtrSrc" :: LeftPar :: Token.String url :: Comma :: Token.String ref :: Comma :: u @ [RightPar]
    | Remove oid ->
	Term "Remove" :: LeftPar :: Nat oid :: [RightPar]
    | RemoveSrc (src,ref) ->
	Term "RemoveSrc" :: LeftPar :: Token.String src :: Comma :: Token.String ref :: [RightPar]
    | Axiom (f,g,url) ->
	Term "Axiom" :: LeftPar :: f @ Comma :: g @ Comma :: Token.String url :: [RightPar]
    | Show f ->
	Term "Show" :: LeftPar :: f @ [RightPar]
    | Hide f ->
	Term "Hide" :: LeftPar :: f @ [RightPar]
    | InitRuleIntr (p,c) ->
	Term "InitRuleIntr" :: LeftPar :: p @ Comma :: c @ [RightPar]
    | InitRuleExtr (p,c) ->
	Term "InitRuleExtr" :: LeftPar :: p @ Comma :: c @ [RightPar]
    | InitRuleRem (p) ->
	Term "InitRuleRem" :: LeftPar :: p @ [RightPar]
    | Action (name,cmd,query) ->
	Term "Action" :: LeftPar :: Token.String name :: Comma :: Token.String cmd :: Comma :: query @ [RightPar]


let oid_update : Logcache.oid -> update -> update =
  (** De-anonymise an update (oid=0) with a given oid. *)
  fun oid -> function
    | Remove 0 -> Remove oid
    | MoveIntr (0,e) -> MoveIntr (oid,e)
    | MoveExtr (0,e) -> MoveExtr (oid,e)
    | up -> up

let mime_match prefix mime = Str.string_match (Str.regexp_string prefix) mime 0

let localize_path : string -> string =
  fun path ->
    if Sys.os_type = "Cygwin"
    then Common.cyg2win path
    else path


(**  The signature that sources must conform to. *)
module type T =
  sig
    type ref = string (** Type of object references in sources. *)

    type t  (** Type of sources. *)

    val parts : t -> ref LSet.t
	(** [parts src] returns the set of all object references generated by the source [src]. *)
(*    val oid : t -> ref -> Logcache.oid*)
        (** [oid src ref] returns the object id of the object identified by reference [ref] in source [src]. *)

    val preview : t -> ref -> preview
	(** [preview src ref] returns a preview of the content of object [ref] generated from source [src]. *)
    val action_args : t -> ref -> action_arg list
	(** [action_args src ref] returns a list of arguments for actions. *)

    val add : t -> ref -> Syntax.t_list -> Syntax.t_list
	(** [add src ref d] changes object [ref] in source [ref] by adding descriptor [d], and returns its new intrinsic description.
	 @raise Invalid_argument when operation is not possible *)
    val sub : t -> ref -> Syntax.t_list -> Syntax.t_list
	(** [sub src ref d] changes object [ref] in source [ref] by removing descriptor [d], and returns its new intrinsic description.
	 @raise Invalid_argument when operation is not possible *)
    val replace : t -> ref -> Syntax.t_list -> Syntax.t_list -> Syntax.t_list
	(** [replace src ref d d'] changes object [ref] in source [ref] by replacing descriptor [d] by descriptor [d'], and returns its new intrinsic description.
	 @raise Invalid_argument when operation is not possible *)

    val init : t -> src_option list -> update list
	(** [init src options] returns a list of updates to be applied at the initialization of the source [src]. *)
    val update : t -> src_option list -> update list
	(** [update src options] returns a list of updates to be applied when the source [src] is updated. *)
    val final : t -> update list  (* get final updates when source is removed *)
	(** [final src] returns a list of updates to be applied when the source [src] is removed. *)
    val move : t -> url -> update list
	(** [move src url] relocates the source [src] at [url]. *)

    val make : url -> t  (* make a new source from an url *)
    val url : t -> url  (* get the url of a source *)
  end

(** A partial default implementation of sources. *)
module Default =
  struct
    type ref = string

    let preview s _ = raise (Invalid_argument "Source.Default.preview: undefined.")

    let action_args s r = []

    let add s r d = raise (Invalid_argument "Src.add: not available")

    let sub s r d = raise (Invalid_argument "Src.sub: not available")

    let replace s r d d' = raise (Invalid_argument "Src.replace: not available")

    let update s _ = []

  end

(*
(** Examples of actions. *)

let action_xemacs = Action ("Open $(text) with XEmacs", "xemacs \"$(text)\" &", Syntax.list_of_string "all")

let action_xemacs_forward = Action ("Open $(text) with XEmacs at $(pattern)", "xemacs -eval '(progn (find-file \"$(text)\") (search-forward \"$(pattern)\"))' &", Syntax.list_of_string "all")

let action_mozilla = Action ("Browse $(url) with Mozilla", "mozilla -remote \"openurl($(url))\"", Syntax.list_of_string "all")

let action_acroread = Action ("Open $(application/pdf) with Acroread", "acroread \"$(application/pdf)\" &", Syntax.list_of_string "all")

let action_gqview = Action ("View $(image) with GQview", "gqview \"$(image)\" &", Syntax.list_of_string "all")

let action_xmms = Action ("Play $(audio/mp3) with XMMS", "xmms \"$(audio/mp3)\" &", Syntax.list_of_string "all")

let action_irfanview = Action ("View $(image) with Irfan", "/cygdrive/c/Program\\ Files/IrfanView/i_view32.exe \"$(image)\" &", Syntax.list_of_string "all")

let action_winamp = Action ("Play $(audio/mp3) with Winamp", "/cygdrive/c/Program\\ Files/Winamp/winamp.exe \"$(audio/mp3)\" &", Syntax.list_of_string "all")
*)
