(** Source: CSV files, with first line giving attribute names. *)

open Source

module Param =
  struct
    let filter url = Filename.check_suffix url ".csv"

    let may_have_changed_since = Url.may_have_changed_since

    let split line =
      let len = String.length line in
      let name, rev_fields, _ =
	Common.fold_while
	  (fun (name,fs,pos) ->
	    if pos >= len then
	      None
	    else if Str.string_match (Str.regexp "\\(\"[^\"]*\"\\)+") line pos then
	      let pos' = Str.match_end () + 1 in
	      let s = Str.matched_string line in
	      let f = Str.global_replace (Str.regexp "\"\"") "\"" (String.sub s 1 (String.length s - 2)) in
	      Some ((if name="" then f else name), f::fs, pos')
	    else if Str.string_match (Str.regexp "\\('[^']*'\\)+") line pos then
	      let pos' = Str.match_end () + 1 in
	      let s = Str.matched_string line in
	      let f = Str.global_replace (Str.regexp "''") "'" (String.sub s 1 (String.length s - 2)) in
	      Some ((if name="" then f else name), f::fs, pos')
	    else if Str.string_match (Str.regexp "[^,;\"]*") line pos then
	      let pos' = Str.match_end () +1 in
	      let f = Str.matched_string line in
	      Some ((if name="" then f else name), f::fs, pos')
	    else
	      None)
	  ("", [], 0) in
      name,
      List.rev_map
	(fun s ->
	  try match Syntax.from_string s with parser
	  | [<n = Syntax.parse_num; _ = Stream.empty>] ->
	      ( match n with
	      | Syntax.Int i -> Fields.Verbatim ("= " ^ string_of_float (float_of_int i))
	      | Syntax.Float (f,_) -> Fields.Verbatim ("= " ^ string_of_float f))
	  | [<_ = Stream.empty>] -> Fields.None
	  with _ -> Fields.StringList [s])
(*
	  | [] -> Fields.None
	  | [Token.Nat n] -> Fields.Integer n
	  | [Token.Minus; Token.Nat n] -> Fields.Integer (-n)
	  | [Token.String s] -> Fields.StringList [s]
	  | _ -> Fields.StringList [s])
*)
	rev_fields

    let rec combine : Fields.v list * Fields.v list -> Fields.t = function
      | _, [] -> []
      | [], v::lv -> 
	  if v = Fields.None
	  then combine ([],lv)
	  else ("other",v)::combine ([],lv)
      | Fields.StringList [a]::la, v::lv ->
	  if v = Fields.None
	  then combine (la,lv)
	  else (a,v)::combine (la,lv)
      | _::la, v::lv ->
	  if v = Fields.None
	  then combine (la,lv)
	  else ("other",v)::combine (la,lv)

    let get_objs url proper_parts =
      let lines =
	if proper_parts
	then
	  let ch_in = open_in (Url.path_of_url url) in
	  let lines =
	    List.rev (Common.fold_in_channel
			(fun res line -> try split line::res with _ -> res)
			[]
			ch_in) in
	  close_in ch_in;
	  lines
	else [] in
      let lf_src = Url.fields_of_url "source" [] url in
      let src_pre = Url.preview_of_url url in
      let src_arg = {mime=Url.mime url "text/csv"; role=""; location=url} in
      ("",(lf_src, String src_pre, src_arg::Url.args_of_url url))::
      match lines with
      | [] -> []
      | (_,attrs)::objs ->
	  List.map
	    (fun (name,obj) ->
	      let descr = combine (attrs, obj) in
	      let ref = try Syntax.stringizer (Fields.print (List.hd descr)) with _ -> "" in
	      let pre = name in (* Syntax.stringizer (Fields.prints descr) in *)
	      let action_args = [{mime="pattern"; role=ref; location=ref}; src_arg] in
	      (ref, (("source", Fields.Composed ("part",Fields.None)) :: lf_src @ descr, String pre, action_args)))
	    objs
  end

module Make : Source.T =
  Setobj.Make(Param)

(*
    (struct
      let suffix = ["csv"]
      let file_attributes = []
      open Token
      let preview path ref pre = Source.String pre
    let main_arg path = {mime="text/csv"; role=""; location=path}
  end)
*)
