(** Source: headers of JPEG images. *)

open Source

module Param =
  struct
    let filter url = List.exists (Filename.check_suffix (String.lowercase url)) [".jpg"; ".jpeg"]

    let may_have_changed_since = Url.may_have_changed_since

    open Fields

    let get_obj url =
	let field v = ("exif", v) in
	let months = [1,"jan"; 2,"feb"; 3,"mar"; 4, "apr"; 5,"may"; 6,"jun"; 7,"jul"; 8,"aug"; 9,"sep"; 10,"oct"; 11,"nov"; 12,"dec"] in
	let lf =
	  try
	    if Sys.os_type <> "Unix" then raise Not_found; (* problems with jhead on Windows/Cygwin *)
	    Common.pipe_command
	      ("jhead -se \"" ^ Url.path_of_url url ^ "\"")
	      (fun ch_exif ->
		Common.fold_in_channel
		  (fun res line ->
		    match Str.split (Str.regexp "[ \t:]+") line with
		    | "Camera"::"make"::l ->
			field (Composed ("camera", Composed ("make", StringList [String.concat " " l])))::res
		    | "Camera"::"model"::l ->
			field (Composed ("camera", Composed ("model", StringList [String.concat " " l])))::res
		    | "Date/Time"::year::month::day::h::m::s::_ ->
			(try
			  field (Date (int_of_string day, List.assoc (int_of_string month) months, int_of_string year))::
			  field (Time (int_of_string h, int_of_string m, int_of_string s))::
			  res
			with _ -> res)
		    | "Resolution"::width::_::height::_ ->
			(try
			  let w, h = int_of_string width, int_of_string height in
			  let size = w * h in
			  let ratio = float (10 * (max w h) / (min w h)) /. 10. in
			  let orientation = if h > w then "portrait" else "landscape" in
(*		  field (Composed ("image", Composed ("width", Integer w))):: *)
(*		  field (Composed ("image", Composed ("height", Integer h))):: *)
			  field (Composed ("image", Composed ("size", Integer size)))::
			  field (Composed ("image", Composed ("orientation", Composed (orientation, None))))::
			  field (Composed ("image", Composed ("ratio", Verbatim ("= " ^ string_of_float ratio))))::
			  res
			with _ -> res)
		    | "Flash"::"used"::l ->
			field (Composed ("exposure", Composed ("flash", StringList [String.concat " " l])))::res
(*
   | "Aperture"::mes::_ ->
		      let a = "= " ^ String.sub mes 2 (String.length mes - 2) in
		      field (Composed ("exposure", Composed ("aperture", Verbatim a)))::res
		  | "Focal"::"length"::mes::_ ->
		      let f = "= " ^ String.sub mes 0 (String.length mes - 2) in
		      field (Composed ("exposure", Composed ("focal_length", Verbatim f)))::res
		  | "CCD"::"width"::mes::_ ->
		      let w = "= " ^ String.sub mes 0 (String.length mes - 2) in
		      field (Composed ("exposure", Composed ("cdd_width", Verbatim w)))::res
*)
		    | "Exposure"::"time"::mes::_ ->
			let t = "= " ^ mes in
			field (Composed ("exposure", Composed ("duration", Verbatim t)))::res
		    | "Exposure"::l ->
			field (Composed ("exposure", Composed ("mode", StringList [String.concat " " l])))::res
		    | _ -> res)
		  []
		  ch_exif)
	  with _ -> [] in
	let lf_src = Url.fields_of_url "source" [] url in
	let pre = Picture (url,Url.preview_of_url url) in
	let args = {mime=Url.mime url "image/jpeg"; role="photo"; location=url}::Url.args_of_url url in
	let lf, args =
	  let url_mov = Filename.chop_extension url ^ ".mov" in
	  if (if Url.is_local url then Url.is_local url_mov else Url.is_on_web url_mov)
	  then ("", TermAsAttr "Video") :: lf, {mime=Url.mime url_mov "video/mov"; role="video"; location=url_mov} :: args
	  else lf, args in
	(lf_src @ lf, pre, args)

  end
module Make : T =
  Sglobj.Make(Param)
