(** Source: Mercury XML files *)

open Source

module Make =
  Setobj.Make
    (struct
      let filter path = Str.string_match (Str.regexp ".*\\.mercury\\.xml$") path 0

      let may_have_changed_since = Url.may_have_changed_since

      open Xml

      type mtype =
	| Pred of mtype list
	| Type of string * mtype list
	| Var of string

      type mmode = string

      type 'a collect = {
	  mutable modul : string;
	  mutable name : string;
	  mutable other : 'a
	}

      type other_predicate = {
	  mutable pred_types : mtype list;
	  mutable pred_modes : mmode list;
	  mutable pred_det : string;
	}
      type collect_predicate = other_predicate collect

      type collect_type = mtype list collect

      type collect_mode = unit collect

      let regexp_ident = Str.regexp "[a-z][_A-Za-z0-9]*$"

      let qualif modul name =
	let s =
	  if modul = "builtin" || modul = name
	  then name
	  else String.concat "__" [modul; name] in
	if Str.string_match regexp_ident s 0
	then s
	else "'" ^ s ^ "'"

      let rec string_of_type = function
	| Pred args ->
	    "pred(" ^ String.concat "," (List.map (fun arg -> "x " ^ string_of_type arg) args) ^ ", is x)"
	| Type (c, args) ->
	    if args = []
	    then c
	    else c ^ "(" ^ String.concat "," (List.map string_of_type args) ^ ")"
	| Var v -> v

      let signature is_fun mtypes mmodes det =
	let l =
	  List.map2
	    (fun t m -> m ^ " " ^ string_of_type t)
	    mtypes mmodes in
	let l = l @ ["is " ^ det] in
	let l = if is_fun then "fun" :: l else l in
	let res = "pred(" ^ String.concat ", " l ^ ")" in
	print_endline res;
	res

      let ignore_xml = function
	| Element (tag, _, _) -> Printf.printf "Ignored tag: %s\n" tag; flush stdout
	| PCData s -> Printf.printf "Ignored PCData: %s\n" s; flush stdout

      let rec parse : xml -> Setobj.descr list = function
	| Element ("module", _, l) ->
	    parse_module l
	| _ -> failwith "error in parse"

      and parse_module : xml list -> Setobj.descr list = function
	| Element ("preds", _, preds)::l ->
	    let objs = parse_preds preds in
	    objs @ parse_module l
	| other::l ->
	    ignore_xml other;
	    parse_module l
	| [] ->
	    print_endline "End of parsing";
	    []

      and parse_preds : xml list -> Setobj.descr list = function
	| Element (tag, _, props)::l when tag = "predicate" || tag = "function" ->
	    begin try
	      let pred = {modul = ""; name = ""; other = {pred_types = []; pred_modes = []; pred_det = ""}} in
	      let () = parse_predicate pred props in
	      let qualif_name = qualif pred.modul pred.name in
	      let obj =
		let ref = qualif_name ^ "__pred" in
		let lf = [
	          "qualified", Fields.Composed ("module", Fields.StringList [pred.modul]);
		  "qualified", Fields.Composed ("name", Fields.StringList [pred.name]);
	          "mercury", Fields.Verbatim (signature
						(tag = "function")
						pred.other.pred_types
						pred.other.pred_modes
						pred.other.pred_det)] in
		let pre = Source.String ("pred " ^ qualif_name) in
		let action_args =  [] in
		ref, (lf, pre, action_args) in
	      obj :: parse_preds l
	    with e ->
	      print_endline (Printexc.to_string e);
	      parse_preds l end
	| other::l ->
	    ignore_xml other;
	    parse_preds l
	| [] -> []

      and parse_predicate (pred : collect_predicate) : xml list -> unit = function
	| Element ("qualified", _, args)::l ->
	    let modul, name = parse_qualified args in
Printf.printf "procedure: %s__%s\n" modul name; flush stdout;
	    pred.modul <- modul;
	    pred.name <- name;
	    parse_predicate pred l
	| Element ("pred_types", _, types)::l ->
	    let t = parse_types types in
	    pred.other.pred_types <- t;
	    parse_predicate pred l
	| Element ("pred_modes", _, [Element ("pred_mode", _, [Element ("modes", _, modes); Element ("determinism", _, [d])])])::l ->
	    let m = parse_modes modes in
	    pred.other.pred_modes <- m;
	    pred.other.pred_det <- pcdata d;
	    parse_predicate pred l
	| other::l ->
	    ignore_xml other;
	    parse_predicate pred l
	| [] -> ()

      and parse_qualified : xml list -> string * string = function
	| [Element ("module", _, [d1]); Element ("name", _, [d2])] ->
	    let modul = pcdata d1 in
	    let name = pcdata d2 in
	    modul, name
	| _ -> failwith "error in parse_qualified"

      and parse_types : xml list -> mtype list = function
	| Element ("type_variable", _, [d])::l ->
	    let mt = Var (pcdata d) in
	    mt :: parse_types l
	| Element ("type", _, typ)::l ->
	    let t = {modul = ""; name = ""; other = []} in
	    let () = parse_type t typ in
	    let mt = Type (qualif t.modul t.name, t.other) in
	    mt :: parse_types l
	| Element ("tuple", _, [Element ("tuple_types", _, types)])::l ->
	    let t = parse_types types in
	    let mt = Type (qualif "builtin" "tuple", t) in
	    mt :: parse_types l
	| Element ("higher_order_type", _, type_args)::l ->
	    let t = parse_higher_order_type type_args in
	    let mt = Pred t in
	    mt :: parse_types l
	| Element (tag, [], [])::l ->
	    let mt = Type (tag, []) in
	    mt :: parse_types l
	| other::l -> ignore_xml other; assert false
	| [] -> []

      and parse_higher_order_type : xml list -> mtype list = function
	| Element ("higher_order_type_args", _, types)::l ->
	    let mts = parse_types types in
	    mts @ parse_higher_order_type l
	| Element ("return_type", _, types)::l ->
	    let mts = parse_types types in
	    mts @ parse_higher_order_type l
	| other::l -> ignore_xml other; assert false
	| [] -> []

      and parse_type (typ : collect_type) : xml list -> unit = function
	| Element ("qualified", _, qualif)::l ->
	    let modul, name = parse_qualified qualif in
	    typ.modul <- modul;
	    typ.name <- name;
	    parse_type typ l
	| Element ("type_args", _, args)::l ->
	    typ.other <- parse_types args;
	    parse_type typ l
	| other::l ->
	    ignore_xml other;
	    parse_type typ l
	| [] -> ()

      and parse_modes : xml list -> mmode list = function
	| Element ("user_defined_mode", _, mode)::l ->
	    let m = {modul = ""; name = ""; other = ()} in
	    let () = parse_mode m mode in
	    let mm = (qualif m.modul m.name) in 
	    mm :: parse_modes l
	| Element ("inst_to_inst", _, _)::l ->
	    let mm = "custom" in
	    mm :: parse_modes l
	| other::l -> ignore_xml other; assert false
	| [] -> []

      and parse_mode (m : collect_mode) : xml list -> unit = function
	| Element ("qualified", _, qualif)::l ->
	    let modul, name = parse_qualified qualif in
	    m.modul <- modul;
	    m.name <- name;
	    parse_mode m l
	| other::l ->
	    ignore_xml other;
	    parse_mode m l
	| [] -> ()


      let get_objs url proper_parts =
	print_endline url;
	let file = Url.path_of_url url in
	if not proper_parts
	then []
	else try
	  let xml = Xml.parse_file file in
	  parse xml
	with
	| Xml.File_not_found s -> print_endline s; []
	| Xml.Error e -> print_endline (Xml.error e); []
	| e -> print_endline (Printexc.to_string e); []

    end)


