(** Source: directories as sources for more sources. *)

module Make
(Param : sig
  val filter : Source.url -> bool
  val may_have_changed_since : Source.url -> float -> bool
  val get_srcs : Source.url -> bool -> (Source.url -> bool) -> Source.url list
  val move_src : Source.url -> Source.url -> Source.url
      (* new url, old src -> new src *)
end) : Source.T =
  struct
    include Source.Default

    open Token

    type t = {
	mutable url : Source.url;
	mutable options : Source.src_option list;
	mutable last_update : float;
	mutable srcs : Source.url LSet.t (* set of generated source urls, even those not existing anymore *)
      }

    let get_srcs url options =
      Param.get_srcs
	url
	(List.mem `Recursive options)
	(List.fold_left (fun res -> function
	  | `Suffixes l ->
	      let l1 = List.map (fun suf -> "." ^ suf) l in
	      (fun src -> List.exists (Filename.check_suffix src) l1)
	  | _ -> res)
	   (fun src -> true) options)

    let make url =
      if Param.filter url
      then {
	url = url;
	options = [];
	last_update = 0.;
	srcs = LSet.empty ()
      }
      else raise (Invalid_argument "Setsrc.make: invalid url")

    let url s = s.url

    let eq s1 s2 = s1.url = s2.url

    let parts s = []

    let init s options =
      s.options <- options;
      s.last_update <- Unix.time ();
      try
	let srcs = get_srcs s.url options in
	s.srcs <- srcs;
	List.map
	  (fun src -> Source.Init (s.url, src, options))
	  srcs
      with _ -> []

    let update s options =
      if s.options = options && not (Param.may_have_changed_since s.url s.last_update)
      then begin
	s.last_update <- Unix.time ();
	[] end
      else begin
	s.options <- options;
	s.last_update <- Unix.time ();
	try
	  let srcs = get_srcs s.url options in
	  let new_srcs = LSet.subtract srcs s.srcs in
	  s.srcs <- LSet.union s.srcs srcs;
	  (List.map
	     (fun src -> Source.Init (s.url, src, options))
	     new_srcs)
	    (* we don't update existing parts as they may ave been finalized *)
	with _ -> [] (* in case directory does not exist any more *)
      end

    let final s = []

    let move s new_url =
      s.url <- new_url;
      let new_srcs, updates =
	List.fold_left
	  (fun (new_srcs, updates) src ->
	    let new_src = Param.move_src s.url src in
	    new_src::new_srcs, Source.Move(src,new_src)::updates)
	  ([], [])
	  s.srcs in
      s.srcs <- new_srcs;
      updates

  end
