(** Source: TNT programs *)

open Source

module Make =
  Setobj.Make
    (struct
      let filter path = Filename.basename path = "tnt.xml"

      let may_have_changed_since = Url.may_have_changed_since

      let get_date_time = (* from format like 20071105055000 *)
	let month = [| ""; "jan"; "feb"; "mar"; "apr"; "may"; "jun"; "jul"; "aug"; "sep"; "oct"; "nov"; "dec"|] in
	let i s pos len = int_of_string (String.sub s pos len) in
	fun s ->
	  (i s 6 2, month.(i s 4 2), i s 0 4),
	  (i s 8 2, i s 10 2, i s 12 2)

      let get_duration (h1,m1,s1 as t1) (h2,m2,s2 as t2) =
	let h2 = if t2 < t1 then h2+24 else h2 in
	60 * (h2 - h1) + (m2 - m1)

      let get_string_field ?ref attr elt =
	let s = Xml.pcdata (List.hd (Xml.children elt)) in
	(match ref with None -> () | Some r -> r := s);
	attr, Fields.StringList [s]

      let get_term_field elt =
	let s = Xml.pcdata (List.hd (Xml.children elt)) in
	"", Fields.TermAsAttr s

      let get_int_field attr elt =
	attr, Fields.Integer (int_of_string (Xml.pcdata (List.hd (Xml.children elt))))

      let get_preview channel (d,mon,y) (h,m,_) dur title desc =
	let buf = Buffer.create 1000 in
	Printf.bprintf buf "[%s] %d %s %d %d:%d (%d min)\n%s" channel d mon y h m dur title;
	let length = String.length desc in
	let nb_columns = 50 in
	let pos1 = ref 0 in (* start of line *)
	let pos2 = ref 0 in (* start of next line *)
	let set_pos2 () =
	  pos2 := min (!pos1 + nb_columns) length;
	  while !pos2 <> length && desc.[!pos2 - 1] <> ' ' do decr pos2 done;
	  if !pos2 - !pos1 < nb_columns / 2 then pos2 := min (!pos1 + nb_columns) length in
	set_pos2 ();
	while !pos1 < !pos2 do
	  Buffer.add_char buf '\n';
	  Buffer.add_string buf (String.sub desc !pos1 (!pos2 - !pos1));
	  pos1 := !pos2;
	  set_pos2 ()
	done;
	Buffer.contents buf

      let get_objs url proper_parts =
	let file = Url.path_of_url url in
	let lf_src = Url.fields_of_url "source" [] url in
	let arg_src = {mime=Url.mime url "text/xml"; role="XML"; location=url} in
	let l =
	  if not proper_parts
	  then []
	  else try
	    let channels = Hashtbl.create 31 in
	    let xml = Xml.parse_file file in
	    Xml.fold
	      (fun res elt ->
		match Xml.tag elt with
		| "channel" ->
		    let id = Xml.attrib elt "id" in
		    let name = String.uppercase (Xml.pcdata (List.hd (Xml.children (List.hd (Xml.children elt))))) in
		    Hashtbl.add channels id name;
		    res
		| "programme" ->
		    let date_time = Xml.attrib elt "start" in
		    let date, time = get_date_time date_time in
		    let date2, time2 = get_date_time (Xml.attrib elt "stop") in
		    let duration = get_duration time time2 in
		    let channel_id = Xml.attrib elt "channel" in
		    let channel = Hashtbl.find channels channel_id in (* String.sub raw_channel 0 (String.index raw_channel '.') in *)
		    let title = ref "titre inconnu" in
		    let desc = ref "" in
		    let lf =
		      Xml.fold
			(fun res2 elt2 ->
			  match Xml.tag elt2 with
			  | "title" -> get_string_field ~ref:title "title" elt2 :: res2
			  | "category" -> get_term_field elt2 :: res2
			  | "desc" | "sub-title" -> get_string_field ~ref:desc "desc" elt2 :: res2
			  | "date" -> get_int_field "year" elt2 :: res2
			  | _ -> res2)
			[ (* "source", Fields.Composed ("part",Fields.None); *)
			 "", Fields.Date date;
			 "", Fields.Time time;
			 "duration", Fields.Integer duration;
			 "", Fields.TermAsAttr channel]
			elt in
		    let ref = date_time ^ channel in
		    let pre = Source.String (try get_preview channel date time duration !title !desc with _ -> "") in
		    let args = [arg_src] in
		    (ref, ((* lf_src @ *) lf, pre, args)) :: res
		| _ -> res)
	      []
	      xml
	  with
	  | Xml.File_not_found s -> print_endline s; []
	  | Xml.Error e -> print_endline (Xml.error e); []
	in
	let url_pre = Url.preview_of_url url in
	(* ("",(lf_src, String url_pre, arg_src :: Url.args_of_url url)) :: *) l
    end)
