(** Well: BibTeX bibliography files. *)

module Make : Well.T =
  struct

    type t = {
	mutable path : string; (* path of the produced text file *)
	mutable order : string; (* name of the attribute used for ordering bibitems *)
      }

    let rec get_part ref0 lexbuf =
      try
	while Bibtex_l.token lexbuf <> Bibtex_p.token_start do () done;
	let ref, (lf, pre, action_args) = Bibtex_p.start Bibtex_l.token lexbuf in
	if ref = ref0
	then ref, lf
      else get_part ref0 lexbuf
      with
      | Parsing.Parse_error -> get_part ref0 lexbuf
      | Bibtex_l.Eof -> (ref0,Fields.init)
	    
    let get_bibitem url ref =
      try
	let lexbuf = Lexing.from_channel (open_in (Url.path_of_url url)) in
	get_part ref lexbuf
      with _ ->
	(ref,Fields.init)
	  
    let get_bibitems cs =
      Common.mapfilter
	(fun args ->
	  try
	    let url = (List.find (fun arg -> Source.mime_match "text/bibtex" arg.Source.mime) args).Source.location in
	    let ref = (List.find (fun arg -> Source.mime_match "ref" arg.Source.mime) args).Source.location in
	    Some (get_bibitem url ref)
	  with Not_found ->
	    None)
	cs
	
    let string_of_field = function
      | Fields.None -> ""
      | Fields.TermAsAttr s -> s
      | Fields.StringList ls -> String.concat ", " ls
      | Fields.Integer n -> string_of_int n
      | Fields.Date (d,m,y) -> string_of_int d ^ " " ^ m ^ " " ^ string_of_int y
      | Fields.Time (h,m,s) -> String.concat ":" (List.map string_of_int [h;m;s])
      | Fields.Composed (b,v) -> raise Not_found
      | Fields.Verbatim s -> s

    let output_bibitem ch (ref,lf) =
      let ref, typ, kwds, lf =
	List.fold_left
	  (fun (ref, typ, kwds, lf) ->
	    function
	      | ("ref", Fields.StringList (r::_)) -> (r, typ, kwds, lf)
	      | ("type", Fields.StringList (t::_)) -> (ref, t, kwds, lf)
	      | ("authors",v) -> (ref, typ, kwds, ("author",v)::lf)
	      | ("author",_) -> (ref, typ, kwds, lf)
	      | ("keyword", Fields.StringList (k::_)) -> (ref, typ, (if kwds = "" then k else k ^ ", " ^ kwds), lf)
	      | (a,v) -> (ref, typ, kwds, (a,v)::lf))
	  ("ref", "type", "", [])
	  lf in
      let lf =
	if kwds = ""
	then lf
	else (("keywords", Fields.StringList [kwds]) :: lf) in
(*
      let typ = match List.assoc "type" lf with Fields.StringList (t::_) -> t | _ -> assert false in
      let ref = match List.assoc "ref" lf with Fields.StringList (r::_) -> r | _ -> assert false in
*)
      output_string ch "@"; output_string ch typ;
      output_string ch "{"; output_string ch ref; output_string ch ",\n";
      List.iter
	(fun (a,v) ->
	  try
	    let sv = string_of_field v in
	    output_string ch "    "; output_string ch a;
	    output_string ch " = {"; output_string ch sv;
	    output_string ch "},\n"
	  with _ -> ())
	(List.rev (List.remove_assoc "ref" (List.remove_assoc "type" lf)));
      output_string ch "}\n\n"
	
    let output_bibtex ch order bibitems =
      let key lf =
	try string_of_field (List.assoc order lf)
	with Not_found -> "" in
      List.iter 
	(fun (ref,lf) ->
	  output_bibitem ch (ref,lf))
	(List.sort
	   (fun (r1,l1) (r2,l2) -> compare (key l1) (key l2))
	   bibitems)
	

    let kinds cs =
      if List.exists
	  (fun args ->
	    List.exists (fun arg -> Source.mime_match "text/bibtex" arg.Source.mime) args &&
	    List.exists (fun arg -> Source.mime_match "ref" arg.Source.mime) args)
	  cs
      then [("bib", "BibTeX file")]
      else []
	  
    let action_args w = [
      {Source.mime = "text/bibtex"; Source.role = "selected bibitems"; Source.location = w.path}
    ]
	
    let eq w1 w2 = w1.path = w2.path
	
    open Token
      
    let make url =
      if Filename.check_suffix url ".bib"
      then { path = url; order = "ref"}
      else raise (Invalid_argument "Wbibtex.make: invalid path")

    let url w = w.path
	
    let init w cs = (* raise Sys_error *)
      let ch = open_out w.path in
      output_bibtex ch w.order (get_bibitems cs);
      close_out ch
	
    let final w _cs = (* raise Sys_error *)
      Sys.remove w.path
	
    let update w cs = (* raise Sys_error *)
      let ch = open_out w.path in
      output_bibtex ch w.order (get_bibitems cs);
      close_out ch
	
    let move w new_url =
      Sys.rename w.path new_url;
      w.path <- new_url

  end
